# -*- coding: utf-8 -*-
"""
Created on Fri Jan  3 11:19:50 2020

@author: icruicks
"""

import igraph as ig, leidenalg as la, numpy as np
from sklearn.utils.validation import check_symmetric
from scipy.sparse import issparse, csr_matrix
from sklearn.metrics import adjusted_mutual_info_score as AMI
from scipy.special import comb
from itertools import product


class Leiden_Unimodal:
    
    def __init__(self, obj_type='RB_Mod', resolution=1.0, n_iterations =-1):
        obj_types = {'CPM': la.CPMVertexPartition, 
                     'RBER': la.RBERVertexPartition, 
                     'RB_Mod': la.RBConfigurationVertexPartition,
                     'Mod': la.ModularityVertexPartition,
                     'Surprise': la.SurpriseVertexPartition
                         }
        self.obj_type = obj_type
        self.obj_func = obj_types[obj_type]
        self.resolution = resolution
        self.n_iterations = n_iterations
        
        
    def fit_transform(self, graph):
        if type(graph) is ig.Graph:
            G =graph
        else:
            G = self._scipy_to_igraph(graph)
            
        if self.obj_type in ['CPM', 'RBER', 'RB_Mod']:
            partition = la.find_partition(G, self.obj_func, n_iterations=self.n_iterations,
                                            resolution_parameter=self.resolution)
        else:
            partition = la.find_partition(G, self.obj_func, n_iterations=self.iterations)
        
        self.modularity_ = partition.quality()
        self.labels_ = np.array(partition.membership)
        return self.labels_
        
        
    def _scipy_to_igraph(self, matrix):
        # matrix.eliminate_zeros()
        sources, targets = matrix.nonzero()
        weights = matrix[sources, targets]
        graph = ig.Graph(n=matrix.shape[0], edges=list(zip(sources, targets)), directed=True, edge_attrs={'weight': weights})
        
        try:
            check_symmetric(matrix, raise_exception=True)
            graph = graph.as_undirected()
        except ValueError:
            pass
        
        return graph
    
    
class Leiden_Multiplex:
    
    def __init__(self, obj_types=None, resolutions=None, modal_weights=None, n_iterations=-1):
        
        self.obj_types = obj_types
        self.resolutions = resolutions
        self.modal_weights = modal_weights
        self.n_iterations = n_iterations
        
    def fit_transform(self, graphs):
        obj_table = {'CPM': la.CPMVertexPartition, 
              'RBER': la.RBERVertexPartition, 
              'RB_Mod': la.RBConfigurationVertexPartition,
              'Mod': la.ModularityVertexPartition,
              'Surprise': la.SurpriseVertexPartition
                  }
        
        G=[]
        for graph in graphs:
            if type(graph) is ig.Graph:
                G.append(graph)
            else:
                G.append(self._scipy_to_igraph(graph))
        
        optimiser = la.Optimiser()
        partitions = []
        for i in range(len(G)):
            if self.obj_types is None:
                partitions.append(la.RBConfigurationVertexPartition(G[i], resolution_parameter=1.0))
            elif self.resolutions is None:
                obj = obj_table[self.obj_types[i]]
                partitions.append(obj(G[i]))
            else:
                obj = obj_table[self.obj_types[i]]
                partitions.append(obj(G[i], resolution_parameter=self.resolutions[i]))
       
        if self.modal_weights is None:
            diff = optimiser.optimise_partition_multiplex(partitions, n_iterations=self.n_iterations)
        else:
            diff = optimiser.optimise_partition_multiplex(partitions, layer_weights = self.modal_weights, n_iterations=self.n_iterations)
            
        self.modularities = [part.modularity for part in partitions]
        self.labels_ = np.array(partitions[0].membership)
        return self.labels_
        
        
    def _scipy_to_igraph(self, matrix):
        matrix.eliminate_zeros()
        sources, targets = matrix.nonzero()
        weights = matrix[sources, targets]
        graph = ig.Graph(n=matrix.shape[0], edges=list(zip(sources, targets)), directed=True, edge_attrs={'weight': weights})
        
        try:
            check_symmetric(matrix, raise_exception=True)
            graph = graph.as_undirected()
        except ValueError:
            pass
        
        return graph
    
    
class MVMC:
    
    def __init__(self, n_iterations=-1, max_clusterings=20, 
                 resolution_tol=1e-2, weight_tol=1e-2, verbose=False):
        
        self.n_iterations = n_iterations
        self.max_clusterings = max_clusterings
        self.resolution_tol = resolution_tol
        self.weight_tol = weight_tol
        self.verbose = verbose
        
    def fit_transform(self, graphs):
        G=[]
        for graph in graphs:
            if type(graph) is ig.Graph:
                G.append(graph)
            else:
                G.append(self._scipy_to_igraph(graph))
                
        if self.verbose:
            for i in range(len(G)):
                print("View Graph {}: num_nodes: {}, num_edges: {}, directed: {}, num_components: {}, num_isolates: {}"
                      .format(i, G[i].vcount(), G[i].ecount(), G[i].is_directed(), 
                              len(G[i].components(mode='WEAK').sizes()), G[i].components(mode='WEAK').sizes().count(1)))
        
        self.weights = []
        self.resolutions =[]
        self.best_modularity =-np.inf
        self.best_clustering = None
        self.best_resolutions = None
        self.best_weights = None
        self.modularities =[]
        self.clusterings =[]
        self.final_iteration = 0
        self.best_iteration = 0
        
        weights = [1]*len(G)
        resolutions =[1]*len(G)
        
        for iterate in range(self.max_clusterings):
            partitions = []
            for i in range(len(G)):
                partitions.append(la.RBConfigurationVertexPartition(G[i], resolution_parameter=resolutions[i]))
                
            optimiser = la.Optimiser()
            diff = optimiser.optimise_partition_multiplex(partitions, layer_weights = weights, n_iterations=self.n_iterations)
            self.clusterings.append(np.array(partitions[0].membership))
            self.modularities.append([part.quality()/(part.graph.ecount() if part.graph.is_directed() else 2*part.graph.ecount()) 
                                      for part in partitions])
            self.weights.append(weights.copy())
            self.resolutions.append(resolutions.copy())
            self.final_iteration +=1
            
            
            if self.verbose:
                print("--------")
                print("Iteration: {} \n Modularities: {} \n Resolutions: {} \n Weights: {}"
                      .format(self.final_iteration, self.modularities[-1], resolutions, weights))
            
            # if np.sum(np.array(self.weights[-1]) * np.array(self.modularities[-1])) > self.best_modularity:
            self.best_clustering = self.clusterings[-1]
            self.best_modularity = np.sum(np.array(self.weights[-1]) * np.array(self.modularities[-1]))
            self.best_resolutions = self.resolutions[-1]
            self.best_weights = self.weights[-1]
            self.best_iteration = self.final_iteration
                
            theta_in, theta_out = self._calculate_edge_probabilities(G)
            for i in range(len(G)):
                resolutions[i] = (theta_in[i] - theta_out[i])/ (np.log(theta_in[i]) - np.log(theta_out[i]))
                weights[i] = (np.log(theta_in[i]) - np.log(theta_out[i]))/(np.mean([np.log(theta_in[j]) - np.log(theta_out[j]) for j in range(len(G))]))

                
            if (np.all(np.abs(np.array(self.resolutions[-1])-np.array(resolutions)) <= self.resolution_tol)
                and np.all(np.abs(np.array(self.weights[-1])-np.array(weights)) <= self.resolution_tol)):
                break
        else:
            best_iteration = np.argmax([np.sum(np.array(self.weights[i]) * np.array(self.modularities[i]))
                                        for i in range(len(self.modularities))])
            self.best_clustering = self.clusterings[best_iteration]
            self.best_modularity = np.sum(np.array(self.weights[best_iteration]) * np.array(self.modularities[best_iteration]))
            self.best_resolutions = self.resolutions[best_iteration]
            self.best_weights = self.weights[best_iteration]
            self.best_iteration = best_iteration
            
            if self.verbose:
                print("MVMC did not converge, best result found: Iteration: {}, Modularity: {}, Resolutions: {}, Weights: {}"
                      .format(self.best_iteration, self.best_modularity, self.best_resolutions, self.best_weights))


        return self.best_clustering
        
    
    def _scipy_to_igraph(self, matrix):
        matrix.eliminate_zeros()
        sources, targets = matrix.nonzero()
        weights = list(matrix.data)
        graph = ig.Graph(n=matrix.shape[0], edges=list(zip(sources, targets)), directed=True, edge_attrs={'weight': weights})
        
        try:
            check_symmetric(matrix, raise_exception=True)
            graph = graph.as_undirected()
        except ValueError:
            pass
        
        if not graph.is_weighted():
            graph.es['weight'] = [1.0] * graph.ecount()
        
        return graph
    
    
    def _calculate_edge_probabilities(self, G):
        theta_in =[]
        theta_out =[]
        clusters = self.clusterings[-1].copy()
        for i in range(len(G)):
            m_in = 0
            m = sum(e['weight'] for e in G[i].es)
            kappa =[]
            G[i].vs['clusters'] = clusters
            for cluster in np.unique(clusters):
                nodes = G[i].vs.select(clusters_eq=cluster)
                m_in += sum(e['weight'] for e in G[i].subgraph(nodes).es)
                if G[i].is_directed():
                    degree_products = np.outer(np.array(G[i].strength(nodes, mode = 'IN', weights='weight')), 
                                               np.array(G[i].strength(nodes, mode = 'OUT', weights='weight')))
                    np.fill_diagonal(degree_products,0)
                    kappa.append(np.sum(degree_products, dtype=np.int64))
                else:
                    kappa.append(np.sum(np.array(G[i].strength(nodes, weights='weight')), dtype=np.int64)**2)
            
            if G[i].is_directed():
                if m_in <=0:
                    # Case when there are no internal edges; every node in its own  cluster
                    theta_in.append(1/G[i].ecount())
                else:
                    theta_in.append((m_in)/(np.sum(kappa, dtype=np.int64)/(2*m)))
                if m-m_in <=0:
                    # Case when all edges are internal; 1 cluster or a bunch of disconnected clusters
                    theta_out.append(1/G[i].ecount())
                else:
                    theta_out.append((m-m_in)/(m-np.sum(kappa, dtype=np.int64)/(2*m)))
            else:
                if m_in <=0:
                    # Case when there are no internal edges; every node in its own  cluster
                    theta_in.append(1/G[i].ecount())
                else:
                    theta_in.append((m_in)/(np.sum(kappa, dtype=np.int64)/(4*m)))
                if m-m_in <=0:
                    # Case when all edges are internal; 1 cluster or a bunch of disconnected clusters
                    theta_out.append(1/G[i].ecount())
                else:
                    theta_out.append((m-m_in)/(m-np.sum(kappa, dtype=np.int64)/(4*m)))

        return theta_in, theta_out
    
    
    
class MVDMC:
    
    def __init__(self, n_iterations=-1, max_clusterings =20, tol = 0.01,
                 resolutions=[0.0001, 0.001, 0.005, 0.01, 0.05]):
        
        self.n_iterations = n_iterations
        self.possible_resolutions = resolutions
        self.max_clusterings = max_clusterings
        self.possible_resolutions = resolutions
        self.tol = tol
        
    def fit_transform(self, graphs):
        G=[]
        for graph in graphs:
            if type(graph) is ig.Graph:
                G.append(graph)
            else:
                G.append(self._scipy_to_igraph(graph))
        
        self.weights = []
        self.best_modularity =-np.inf
        self.best_clustering = None
        self.best_resolutions = None
        self.best_weights = None
        self.modularities =[]
        self.clusterings =[]
        self.resolutions =[]
        # self.density_ratios =[0]*len(G)
        self.final_iteration = 0
        self.best_iteration = 0
        
        weights = [1]*len(G)
        # resolutions =[0]*len(G)
        
        for resolutions in list(product(self.possible_resolutions, repeat=len(G))):
        # for iterate in range(self.max_clusterings):
        # for iterate in range(len(self.possible_resolutions)):
            partitions = []
            # resolutions = [self.possible_resolutions[iterate]]*len(G)
            for i in range(len(G)):
                partitions.append(la.CPMVertexPartition(G[i], resolution_parameter=resolutions[i]))
                
            optimiser = la.Optimiser()
            diff = optimiser.optimise_partition_multiplex(partitions, layer_weights = weights, n_iterations=self.n_iterations)
            self.clusterings.append(np.array(partitions[0].membership))
            # self.modularities.append([part.quality()/(part.graph.ecount() if part.graph.is_directed() else 2*part.graph.ecount()) 
            #                            for part in partitions])
            self.modularities.append([part.modularity for part in partitions])
            self.weights.append(weights.copy())
            self.resolutions.append(list(resolutions).copy())
            self.final_iteration +=1
            # d_in = self._calculate_cluster_density_ratios(G)
            # self.density_ratios.append([d_in[i] / (len(np.unique(self.clusterings[-1]))/G[i].vcount())**2 for i in range(len(G))])
            
            # if np.sum(self.density_ratios[-1]) > np.sum(self.density_ratios[-2]):
            if np.sum(np.array(self.weights[-1]) * np.array(self.modularities[-1])) > self.best_modularity:
                self.best_clustering = self.clusterings[-1]
                self.best_modularity = np.sum(np.array(self.weights[-1]) * np.array(self.modularities[-1]))
                self.best_resolution = resolutions
                self.best_weights = self.weights[-1]
                self.best_iteration = self.final_iteration
            
            # resolutions = np.mean(np.array([resolutions, d_in]), axis=0)
            
        return self.best_clustering
        
    
    def _scipy_to_igraph(self, matrix):
        matrix.eliminate_zeros()
        sources, targets = matrix.nonzero()
        weights = list(matrix.data)
        graph = ig.Graph(n=matrix.shape[0], edges=list(zip(sources, targets)), directed=True, edge_attrs={'weight': weights})
        
        try:
            check_symmetric(matrix, raise_exception=True)
            graph = graph.as_undirected()
        except ValueError:
            pass
        
        if 'weight' not in graph.es:
            graph.es['weight'] = [1.0] * graph.ecount()
        
        return graph
    
    
    def _calculate_cluster_density_ratios(self, G):
        d_in =[]
        clusters = self.clusterings[-1].copy()
        for i in range(len(G)):
            G[i].vs['clusters'] = clusters
            d_in_g =[]
            for cluster in np.unique(clusters):
                nodes = G[i].vs.select(clusters_eq=cluster)
                G_cluster = G[i].subgraph(nodes)
                m_in = G_cluster.ecount()+1
                n_in = G_cluster.vcount()
                d_in_g.append(m_in/n_in**2)
            d_in.append(np.sum(d_in_g)/len(np.unique(clusters)))
            # d_in.append(np.min(d_in_g))
            
        return d_in
    
    
    
    